from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector

from famsa.core.sequence cimport CSequence
from famsa.tree cimport tree_structure


cdef extern from "tree/NewickParser.h" nogil:

    cdef cppclass NewickParser:
        NewickParser()

        void parse(
            const vector[CSequence]& sequences,
            const string& description,
            tree_structure& guideTree,
        ) except +

        void store(
            const vector[CSequence]& sequences,
            const tree_structure& guideTree,
            string& description,
        ) except +



# coding: utf-8
# cython: language_level=3, linetrace=True
"""Bindings to FAMSA, an algorithm for fast multiple sequence alignments.

References:
    - Deorowicz, S., Debudaj-Grabysz, A., Gudyś, A. (2016)
      *FAMSA: Fast and accurate multiple sequence alignment of huge protein
      families*. Scientific Reports, 6, 33964. :doi:`10.1038/srep33964`.

"""

# --- C imports --------------------------------------------------------------

from cpython cimport Py_buffer
from cpython.memoryview cimport PyMemoryView_FromMemory
from cpython.buffer cimport PyBUF_FORMAT, PyBUF_READ
from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_AS_STRING

from libc.stdint cimport uint32_t
from libc.string cimport memset
from libcpp cimport bool
from libcpp.memory cimport shared_ptr
from libcpp.utility cimport move
from libcpp.vector cimport vector
from libcpp.string cimport string

from famsa.core cimport symbol_t, GAP, GUARD
from famsa.core.params cimport CParams
from famsa.core.sequence cimport CSequence, CGappedSequence
from famsa.msa cimport CFAMSA
from famsa.tree cimport GT, node_t
from famsa.tree.guide_tree cimport GuideTree as CGuideTree
from famsa.tree.newick_parser cimport NewickParser
from famsa.tree.abstract_tree_generator cimport AbstractTreeGenerator
from famsa.utils.memory_monotonic cimport memory_monotonic_safe
# from famsa.utils.log cimport Log, LEVEL_NORMAL, LEVEL_DEBUG, LEVEL_VERBOSE


# --- Python imports ---------------------------------------------------------

import os


# --- Constants --------------------------------------------------------------

cdef memory_monotonic_safe* MMA = new memory_monotonic_safe()

cdef char SYMBOLS[25]
for i, x in enumerate(b"ARNDCQEGHILKMFPSTWYVBZX*"):
    SYMBOLS[i] = x

# Log.getInstance(LEVEL_NORMAL).enable()
# Log.getInstance(LEVEL_VERBOSE).enable()
# Log.getInstance(LEVEL_DEBUG).enable()

# --- Utils ------------------------------------------------------------------

cdef extern from *:
    """
    void sort_sequences(vector<CSequence>& sequences) {
        std::stable_sort(sequences.begin(), sequences.end(), [](const CSequence& a, const CSequence& b) -> bool {
          return a.length > b.length || (a.length == b.length && std::lexicographical_compare(a.data, a.data + a.data_size, b.data, b.data + b.data_size));
        });
    }
    void shuffle_sequences(vector<CSequence>& sequences, int shuffle) {
        std::mt19937 mt(shuffle);
        std::shuffle(sequences.begin(), sequences.end(), mt);
    }
    """
    void sort_sequences(vector[CSequence]& sequences)
    void shuffle_sequences(vector[CSequence]& sequences, int shuffle)


# --- Classes ----------------------------------------------------------------

cdef class Sequence:
    """A digitized sequence.
    """

    # --- Magic methods ------------------------------------------------------

    def __init__(self, bytes id, bytes sequence):
        self._cseq = move(CSequence(id, sequence, MMA))
        self._shape[0] = self._cseq.length
        assert self._cseq.mma is not NULL

    def __copy__(self):
        return self.copy()

    def __getbuffer__(self, Py_buffer* buffer, int flags):
        if flags & PyBUF_FORMAT:
            buffer.format = b"b"
        else:
            buffer.format = NULL
        buffer.buf = self._cseq.data
        buffer.internal = NULL
        buffer.itemsize = sizeof(symbol_t)
        buffer.len = self._shape[0] * sizeof(symbol_t)
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self._shape
        buffer.suboffsets = NULL
        buffer.strides = NULL

    def __repr__(self):
        cdef str ty = type(self).__name__
        return f"{ty}({self.id}, {self.sequence})"

    # --- Properties ---------------------------------------------------------

    @property
    def id(self):
        """`bytes`: The identifier of the sequence.
        """
        return <bytes> self._cseq.id

    @property
    def sequence(self):
        """`bytes`: The symbols of the sequence as an ASCII string.
        """
        # code from `CSequence::DecodeSequence`
        cdef uint32_t i
        cdef bytes    seq = PyBytes_FromStringAndSize(NULL, self._cseq.length)
        cdef char*    mem = PyBytes_AS_STRING(seq)

        with nogil:
            for i in range(self._cseq.length):
                if self._cseq.data[i] == GUARD:
                    continue
                elif self._cseq.data[i] == GAP:
                    mem[0] = b'-'
                else:
                    mem[0] = SYMBOLS[self._cseq.data[i]]
                mem += 1

        return seq

    # --- Methods ------------------------------------------------------------

    cpdef Sequence copy(self):
        """copy(self)\n--

        Copy the sequence data, and return the copy.

        """
        cdef Sequence seq = Sequence.__new__(Sequence)
        seq._cseq = move(CSequence(self._cseq))
        seq._shape = self._shape
        return seq


cdef class GappedSequence:
    """A gapped sequence, storing a single row in an alignment.

    Internally, gapped sequences are not stored in a single string in FAMSA,
    but using a gap counter for each position. This allows saving a lot of
    memory for alignments containing lots of gaps, but adds some overhead to
    decode the full sequence.

    """

    # --- Properties ---------------------------------------------------------

    @property
    def id(self):
        """`bytes`: The identifier of the gapped sequence.
        """
        return <bytes> self._gseq.id

    @property
    def sequence(self):
        """`bytes`: The symbols of the gapped sequence as an ASCII string.
        """
        # code from `CSequence::DecodeSequence`
        cdef uint32_t i
        cdef char     symbol
        cdef bytes    seq    = PyBytes_FromStringAndSize(NULL, self._gseq.gapped_size)
        cdef char*    mem    = PyBytes_AS_STRING(seq)

        with nogil:

            # starting gaps
            memset(mem, b'-', self._gseq.n_gaps[0])
            mem += self._gseq.n_gaps[0]

            for i in range(1, self._gseq.size+1):
                symbol = SYMBOLS[self._gseq.symbols[i]]
                if not self._gseq.uppercase[i - 1]:
                    symbol += 32

                mem[0] = symbol
                mem += 1

                memset(mem, b'-', self._gseq.n_gaps[i])
                mem += self._gseq.n_gaps[i]

        return seq


cdef class Alignment:
    """An alignment, stored as a list of `GappedSequence` objects.
    """

    # --- Magic methods ------------------------------------------------------

    def __init__(self):
        self._msa.clear() # create an empty alignment

    def __len__(self):
        return self._msa.size()

    def __getitem__(self, ssize_t index):
        cdef GappedSequence gapped
        cdef ssize_t        index_ = index

        if index_ < 0:
            index_ += self._msa.size()
        if index_ < 0 or index_ >= <ssize_t> self._msa.size():
            raise IndexError(index)

        gapped = GappedSequence.__new__(GappedSequence)
        gapped.alignment = self
        gapped._gseq = self._msa[index_]
        return gapped


cdef class Aligner:
    """A single FAMSA aligner.
    """

    # --- Magic methods ------------------------------------------------------

    def __cinit__(self):
        self._params = CParams()
        self._params.verbose_mode = True
        self._params.very_verbose_mode = True
        self._params.n_threads = 1

    def __init__(
        self,
        *,
        int threads=0,
        object guide_tree="sl",
        object tree_heuristic=None,
        int medoid_threshold=0,
        int n_refinements=100,
        bool force_refinement=False,
    ):
        """__init__(self, *, threads=0, guide_tree="sl", tree_heuristic=None, medoid_threshold=0, n_refinements=100, force_refinement=False)\n--

        Create a new aligner with the given configuration.

        Keyword Arguments:
            threads (`int`): The number of threads to use for parallel
                computations. If *0* given (the default), use `os.cpu_count`
                to spawn one thread per CPU on the host machine.
            guide_tree (`str`): The method for building the guide tree.
                Supported values are: ``sl`` for MST+Prim single linkage,
                ``slink`` for SLINK single linkage, ``upgma`` for UPGMA,
                ``nj`` for neighbour joining.
            tree_heuristic (`str` or `None`): The heuristic to use for
                constructing the tree. Supported values are: ``medoid`` for
                medoid trees, ``part`` for part trees, or `None` to disable
                heuristics.
            medoid_threshold (`int`): The minimum number of sequences a
                set must contain for medoid trees to be used, if enabled
                with ``tree_heuristic``.
            n_refinements (`int`): The number of refinement iterations to
                run.
            force_refinement (`bool`): Set to `True` to force refinement
                on sequence sets larger than 1000 sequences.

        """
        if threads == 0:
            self._params.n_threads = os.cpu_count() or 1
        elif threads >= 1:
            self._params.n_threads = threads
        else:
            raise ValueError("`threads` argument must be positive")

        if guide_tree == "sl":
            self._params.gt_method = GT.Method.MST_Prim
        elif guide_tree == "slink":
            self._params.gt_method = GT.Method.SLINK
        elif guide_tree == "upgma":
            self._params.gt_method = GT.Method.UPGMA
        elif guide_tree == "nj":
            self._params.gt_method = GT.Method.NJ
        else:
            raise ValueError(f"Invalid value for `guide_tree` argument: {guide_tree!r}")

        if tree_heuristic is None:
            self._params.gt_heuristic = GT.Heuristic.None
        elif tree_heuristic == 'medoid':
            self._params.gt_heuristic = GT.Heuristic.ClusterTree
        elif tree_heuristic == 'parttree':
            self._params.gt_heuristic = GT.Heuristic.PartTree
        else:
            raise ValueError(f"Invalid value for `tree_heuristic` argument: {tree_heuristic!r})")

        if medoid_threshold >= 0:
            self._params.heuristic_threshold = medoid_threshold
        else:
            raise ValueError("`medoid_threshold` argument must be positive")

        if n_refinements >= 0:
            self._params.n_refinements = n_refinements
        else:
            raise ValueError("`n_refinements` argument must be positive")

        if force_refinement:
            self._params.enable_auto_refinement = False

    # --- Methods ------------------------------------------------------------

    cpdef Alignment align(self, object sequences):
        """align(self, sequences)\n--

        Align sequences together.

        Arguments:
            sequences (iterable of `~pyfamsa.Sequence`): An iterable
                yielding the digitized sequences to align.

        Returns:
            `~pyfamsa.Alignment`: The aligned sequences, in aligned format.

        """
        cdef Sequence                 sequence
        cdef vector[CSequence]        seqvec
        cdef vector[CGappedSequence*] gapvec
        cdef Alignment                alignment = Alignment.__new__(Alignment)

        # create a new aligner
        alignment._famsa = shared_ptr[CFAMSA](new CFAMSA(self._params))
        # copy the aligner input
        for sequence in sequences:
            seqvec.push_back(move(CSequence(sequence._cseq)))

        # align the input and extract the resulting alignment
        with nogil:
            alignment._famsa.get().ComputeMSA(seqvec)
            alignment._famsa.get().GetAlignment(alignment._msa)

        return alignment

    cpdef GuideTree build_tree(self, object sequences):
        """build_tree(self, sequences)\n--

        Build a tree from the given sequences.

        Arguments:
            sequences (iterable of `~pyfamsa.Sequence`): An iterable
                yielding the digitized sequences to build a tree from.

        Returns:
            `~pyfamsa.GuideTree`: The guide tree obtained from the sequences.

        """
        cdef size_t                            i
        cdef Sequence                          sequence
        cdef vector[CSequence]                 seqvec
        cdef vector[CSequence]                 namevec
        cdef vector[CGappedSequence*]          gapvec
        cdef shared_ptr[AbstractTreeGenerator] gen
        cdef CFAMSA*                           famsa    = new CFAMSA(self._params)
        cdef GuideTree                         tree     = GuideTree.__new__(GuideTree)

        # copy the aligner input
        for sequence in sequences:
            seqvec.push_back(move(CSequence(sequence._cseq)))

        # sort or shuffle sequences
        if self._params.shuffle == -1:
            sort_sequences(seqvec)
        else:
            shuffle_sequences(seqvec, self._params.shuffle)

        # set sequence identifiers and record names
        for i in range(seqvec.size()):
            seqvec[i].sequence_no = i
            tree._names.push_back(move(CSequence(seqvec[i].id, string(), NULL)))

        # generate tree
        try:
            with nogil:
                gen = famsa.createTreeGenerator(self._params)
                gen.get().call( seqvec, tree._tree.raw() )
        finally:
            del famsa

        # return the result tree
        return tree


cdef class GuideTree:
    """A guide tree generated from several sequences.

    Guide trees are binary trees, stored as an array which contains the
    two children nodes for each node (or *-1* for leaves). Nodes can be
    iterated upon.

    """

    # --- Magic methods ------------------------------------------------------

    def __len__(self):
        return self._tree.raw().size()

    def __getitem__(self, size_t index):
        cdef GappedSequence gapped
        cdef node_t         node
        cdef ssize_t        index_ = index

        if index_ < 0:
            index_ += self._tree.raw().size()
        if index_ < 0 or index_ >= <ssize_t> self._tree.raw().size():
            raise IndexError(index)

        node = self._tree.raw()[index]
        return node

    # --- Properties ---------------------------------------------------------

    @property
    def names(self):
        """`list` of `bytes`: The name of the leaves in the guide tree.
        """
        return [ <bytes> self._names[i].id for i in range(self._names.size()) ]

    # --- Methods ------------------------------------------------------------

    cpdef bytes dumps(self):
        """dumps(self)\n--

        Dump the tree in Newick format into a `bytes` object.

        Returns:
            `bytes`: The tree in Newick format, stored as an ASCII string.

        """
        cdef string       out
        cdef NewickParser nw

        with nogil:
            nw.store(self._names, self._tree.raw(), out)

        return <bytes> out

    cpdef ssize_t dump(self, object file) except -1:
        """dump(self, file)\n--

        Write the tree into a file in Newick format.

        Arguments:
            file (`str`, `bytes, `os.PathLike` or file-like object): The
                path to a file, or a file-like object open in binary mode.

        """
        cdef string       out
        cdef NewickParser nw
        cdef object       mem
        cdef bytes        path = None

        # accept either a path or a file-like object
        if isinstance(file, (os.PathLike, str)):
            path = os.fsencode(file)
        elif isinstance(file, bytes):
            path = file
        else:
            path = None

        # dump the tree to a C++ string
        with nogil:
            nw.store(self._names, self._tree.raw(), out)

        # use a memoryview to avoid copy from the C++ string before writing
        mem = PyMemoryView_FromMemory(<char*> out.c_str(), out.size(), PyBUF_READ )

        # write the result to the file and return the number of bytes written
        if path is None:
            return file.write(mem)
        else:
            with open(path, "wb") as handle:
                return handle.write(mem)



# noqa: D104

from . import (
    test_aligner,
    test_doctest,
    test_sequence,
)


def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_aligner))
    suite.addTests(loader.loadTestsFromModule(test_doctest))
    suite.addTests(loader.loadTestsFromModule(test_sequence))
    return suite



from libc.stdint cimport int64_t

from famsa.tree cimport tree_structure


cdef extern from "tree/GuideTree.h" nogil:

    cdef cppclass GuideTree:
        GuideTree()

        tree_structure& raw()
        int getSequenceCount()
        int64_t calculateSackinIndex()



from libcpp.vector cimport vector

from famsa.core cimport instruction_set_t
from famsa.core.sequence cimport CSequence
from famsa.tree cimport tree_structure


cdef extern from "tree/AbstractTreeGenerator.h" nogil:

    cdef cppclass AbstractTreeGenerator:
        AbstractTreeGenerator(int n_threads, instruction_set_t instruction_set) except +

        void call "operator()"(vector[CSequence]& sequences, tree_structure& tree) except +



from libcpp cimport bool

cdef extern from "utils/memory_monotonic.h" namespace "refresh" nogil:

    cdef cppclass memory_monotonic_base:
        memory_monotonic_base(size_t block_size, size_t alignment)


    cdef cppclass memory_monotonic_unsafe(memory_monotonic_base):
        memory_monotonic_unsafe()
        memory_monotonic_unsafe(size_t _block_size, size_t _alignment)

        bool deallocation_status() except +
        void* allocate(size_t size) except +
        void deallocate[T](T*& p) except +
        void freeze() except +
        void release() except +
        void release_freezed() except +


    cdef cppclass memory_monotonic_safe(memory_monotonic_base):
        memory_monotonic_safe()
        memory_monotonic_safe(size_t _block_size, size_t _alignment)

        bool deallocation_status() except +
        void* allocate(size_t size) except +
        void deallocate[T](T*& p) except +
        void freeze() except +
        void release() except +
        void release_freezed() except +



import os
import typing

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

GuideTreeMethod = Literal["sl", "slink", "upgma", "nj"]
TreeHeuristicMethod = Literal["medoid", "part"]
Node = typing.Tuple[int, int]

class Sequence:
    def __init__(self, id: bytes, sequence: bytes) -> None: ...
    def __copy__(self) -> Sequence: ...
    def __repr__(self) -> str: ...
    @property
    def id(self) -> bytes: ...
    @property
    def sequence(self) -> bytes: ...
    def copy(self) -> Sequence: ...

class GappedSequence:
    @property
    def id(self) -> bytes: ...
    @property
    def sequence(self) -> bytes: ...

class Alignment(typing.Sequence[GappedSequence]):
    def __init__(self) -> None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> GappedSequence: ...  # type: ignore

class Aligner:
    def __init__(
        self,
        *,
        threads: int = 0,
        guide_tree: GuideTreeMethod = "sl",
        tree_heuristic: typing.Optional[TreeHeuristicMethod] = None,
        medoid_threshold: int = 0,
        n_refinements: int = 100,
        force_refinement: bool = False,
    ) -> None: ...
    def align(self, sequences: typing.Iterable[Sequence]) -> Alignment: ...
    def build_tree(self, sequences: typing.Iterable[Sequence]) -> GuideTree: ...

class GuideTree(typing.Sequence[Node]):
    def __init__(self) -> None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> Node: ...  # type: ignore
    def dumps(self) -> bytes: ...
    def dump(
        self,
        file: typing.Union[
            str, bytes, os.PathLike[bytes], os.PathLike[str], typing.BinaryIO
        ],
    ) -> int: ...



import unittest

from .. import Sequence


class TestSequence(unittest.TestCase):

    def test_sequence_property(self):
        seq = Sequence(b"test", b"MYYK")
        self.assertEqual(seq.sequence, b"MYYK")

    def test_id_property(self):
        seq = Sequence(b"test", b"MYYK")
        self.assertEqual(seq.id, b"test")

    def test_buffer_export(self):
        seq = Sequence(b"test", b"MYYK")
        mem = memoryview(seq)
        self.assertEqual(mem.shape[0], len(seq.sequence))
        self.assertEqual(mem[1], mem[2])
        self.assertTrue(mem.readonly)
        b = mem.tobytes()
        self.assertEqual(len(b), 4)

    def test_buffer_export_empty(self):
        seq = Sequence(b"", b"")
        mem = memoryview(seq)
        self.assertEqual(mem.shape[0], 0)
        self.assertTrue(mem.readonly)
        b = mem.tobytes()
        self.assertEqual(len(b), 0)



# distutils: language = c++
# cython: language_level=3, linetrace=True

# --- C imports --------------------------------------------------------------

from libcpp.memory cimport shared_ptr
from libcpp.vector cimport vector

from famsa.msa cimport CFAMSA
from famsa.core.params cimport CParams
from famsa.core.sequence cimport CSequence, CGappedSequence
from famsa.tree.guide_tree cimport GuideTree as CGuideTree
from famsa.utils.memory_monotonic cimport memory_monotonic_safe

# --- Allocator --------------------------------------------------------------

cdef memory_monotonic_safe* MMA = new memory_monotonic_safe()

# --- Classes ----------------------------------------------------------------

cdef class Sequence:
    cdef          CSequence  _cseq
    cdef readonly Py_ssize_t _shape[1]

    cpdef Sequence copy(self)


cdef class GappedSequence:
    cdef Alignment        alignment
    cdef CGappedSequence* _gseq


cdef class Alignment:
    cdef shared_ptr[CFAMSA]       _famsa
    cdef vector[CGappedSequence*] _msa


cdef class Aligner:
    cdef CParams _params

    cpdef Alignment align(self, object sequences)
    cpdef GuideTree build_tree(self, object sequences)

cdef class GuideTree:
    cdef CGuideTree        _tree
    cdef vector[CSequence] _names

    cpdef bytes dumps(self)
    cpdef ssize_t dump(self, object file) except -1



# coding: utf-8
"""Test doctest contained tests in every file of the module.
"""

import configparser
import doctest
import importlib
import os
import pkgutil
import re
import shutil
import sys
import types
import warnings
from unittest import mock

import pyfamsa


def _load_tests_from_module(tests, module, globs, setUp=None, tearDown=None):
    """Load tests from module, iterating through submodules.
    """
    for attr in (getattr(module, x) for x in dir(module) if not x.startswith("_")):
        if isinstance(attr, types.ModuleType):
            suite = doctest.DocTestSuite(
                attr,
                globs,
                setUp=setUp,
                tearDown=tearDown,
                optionflags=+doctest.ELLIPSIS,
            )
            tests.addTests(suite)
    return tests


def load_tests(loader, tests, ignore):
    """`load_test` function used by unittest to find the doctests.
    """
    _current_cwd = os.getcwd()

    def setUp(self):
        warnings.simplefilter("ignore")
        os.chdir(os.path.realpath(os.path.join(__file__, os.path.pardir, "data")))

    def tearDown(self):
        os.chdir(_current_cwd)
        warnings.simplefilter(warnings.defaultaction)

    # doctests are not compatible with `green`, so we may want to bail out
    # early if `green` is running the tests
    if sys.argv[0].endswith("green"):
        return tests

    # recursively traverse all library submodules and load tests from them
    packages = [None, pyfamsa]

    for pkg in iter(packages.pop, None):
        for (_, subpkgname, subispkg) in pkgutil.walk_packages(pkg.__path__):
            # do not import __main__ module to avoid side effects!
            if subpkgname == "__main__" or subpkgname.startswith("tests"):
                continue
            # import the submodule and add it to the tests
            module = importlib.import_module(".".join([pkg.__name__, subpkgname]))
            globs = dict(
                **module.__dict__
            )
            tests.addTests(
                doctest.DocTestSuite(
                    module,
                    globs=globs,
                    setUp=setUp,
                    tearDown=tearDown,
                    optionflags=+doctest.ELLIPSIS,
                )
            )
            # if the submodule is a package, we need to process its submodules
            # as well, so we add it to the package queue
            if subispkg and subpkgname != "tests":
                packages.append(module)

    return tests



from libcpp.pair cimport pair
from libcpp.vector cimport vector

cdef extern from "tree/TreeDefs.h" nogil:

      ctypedef pair[int, int] node_t
      ctypedef vector[node_t] tree_structure

      cdef cppclass GT:
          pass

cdef extern from "tree/TreeDefs.h" namespace "GT" nogil:
    cdef enum Method:
        SLINK
        MST_Prim
        UPGMA
        UPGMA_modified
        NJ
        chained
        imported

    cdef enum Heuristic:
        None
        PartTree
        ClusterTree



from libcpp cimport bool
from libcpp.memory cimport shared_ptr
from libcpp.vector cimport vector

from famsa.core cimport score_t, instruction_set_t
from famsa.core.params cimport CParams
from famsa.core.sequence cimport CSequence, CGappedSequence
from famsa.tree cimport tree_structure
from famsa.tree.abstract_tree_generator cimport AbstractTreeGenerator

cdef extern from "msa.h" nogil:

    cdef cppclass CFAMSA:
        instruction_set_t instruction_set
        score_t avg_sim

        CFAMSA(CParams& params) except +

        bool ComputeAlignment(tree_structure& guide_tree) except +
        # bool RefineAlignment(CProfile *&profile_to_refine)
        bool GetAlignment(vector[CGappedSequence*]& result) except +
        score_t GetScore()

        # const Statistics& getStatistics()
        # Statistics& getStatistics()

        bool ComputeMSA(vector[CSequence]& sequences)  except +

        shared_ptr[AbstractTreeGenerator] createTreeGenerator(const CParams& params) except +



from libc.stdint cimport int64_t

cdef extern from "core/defs.h" nogil:

    ctypedef int64_t score_t

    cdef enum instruction_set_t:
        none
        sse
        sse2
        sse3
        sse3s
        sse41
        sse42
        avx
        avx2

    ctypedef char symbol_t
    ctypedef int counter_t

    ctypedef unsigned long long bit_vec_t

    const int bv_size
    const int bv_size128
    const int bv_size256

    const score_t infty
    const double cost_cast_factor

    const symbol_t GAP
    const symbol_t GAP_OPEN
    const symbol_t GAP_EXT
    const symbol_t GAP_TERM_EXT
    const symbol_t GAP_TERM_OPEN
    const symbol_t UNKNOWN_SYMBOL

    const size_t NO_SYMBOLS
    const symbol_t GUARD
    const symbol_t NO_AMINOACIDS
    const symbol_t NO_VALID_AMINOACIDS
    const symbol_t NO_AMINOACIDS_AND_GAPS
    const symbol_t NO_AA_SYMBOLS



from libcpp cimport bool

cdef extern from "utils/log.h" nogil:

    cdef cppclass Log:
        void enable()
        void disable()
        bool isEnabled()

        @staticmethod
        Log& getInstance(int level) except +


cdef extern from "utils/log.h" namespace "Log" nogil:
    const int LEVEL_NORMAL
    const int LEVEL_VERBOSE
    const int LEVEL_DEBUG



# noqa: D104
from ._version import __version__  # isort: skip

from . import _famsa
from ._famsa import (
    Aligner,
    Alignment,
    GappedSequence,
    GuideTree,
    Sequence,
)

__doc__ = _famsa.__doc__
__all__ = [
    "Aligner",
    "Alignment",
    "GappedSequence"
    "GuideTree"
    "Sequence",
]

__author__ = "Martin Larralde <martin.larralde@embl.de>"
__license__ = "GPLv3"



from libc.stdint cimport int64_t, uint32_t, uint64_t
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector

from famsa.core cimport score_t, instruction_set_t
from famsa.tree cimport GT

cdef extern from "core/params.h" nogil:

    cdef enum Distance:
        indel_div_lcs
        sqrt_indel_div_lcs
        neg_lcs_div_indel
        neg_lcs_div_minlen
        neg_lcs_div_len_corrected

    cdef cppclass CParams:
        score_t gap_open
        score_t gap_ext
        score_t gap_term_open
        score_t gap_term_ext

        uint32_t scaler_div
        uint32_t scaler_log
        int guided_alignment_radius

        bool enable_gap_rescaling
        bool enable_gap_optimization
        bool enable_total_score_calculation
        bool enable_auto_refinement

        uint32_t n_refinements
        uint32_t thr_refinement
        uint32_t thr_internal_refinement

        GT.Method gt_method
        GT.Heuristic gt_heuristic
        Distance distance
        int heuristic_threshold

        int guide_tree_seed
        int subtree_size
        int sample_size
        float cluster_fraction
        int cluster_iters

        string guide_tree_in_file
        bool export_distances
        bool export_tree
        bool generate_square_matrix
        bool calculate_pid

        bool test_ref_sequences
        uint64_t ref_seq_subtree_size
        string ref_file_name

        int64_t shuffle
        uint32_t n_threads

        bool gzippd_output
        int gzip_level

        instruction_set_t instruction_set

        bool verbose_mode
        bool very_verbose_mode

        string input_file_name
        string output_file_name

        vector[vector[score_t]] score_matrix
        vector[score_t] score_vector

        CParams() except +
        bool parse(int argc, char** argv, bool& showExpert) except +
        void show_usage(bool expert) except +



from libc.stdint cimport uint32_t
from libcpp cimport bool
from libcpp.pair cimport pair
from libcpp.string cimport string
from libcpp.vector cimport vector

from famsa.core cimport bit_vec_t, symbol_t
from famsa.utils.memory_monotonic cimport memory_monotonic_safe


cdef extern from "core/sequence.h" nogil:

    cdef cppclass CSequence:
        uint32_t length
        uint32_t data_size
        symbol_t* data
        bit_vec_t* p_bit_masks
        uint32_t p_bv_len

        int sequence_no
        string id

        memory_monotonic_safe* mma

        vector[bool] uppercase

        CSequence()
        CSequence(const string& id, const string& seq, memory_monotonic_safe* mma) except +
        CSequence(CSequence&& x)

        void DataResize(uint32_t new_size, symbol_t new_symbol) except +

        void ComputeBitMasks() except +
        void ReleaseBitMasks() except +
        string DecodeSequence() except +

    cdef struct CSequenceView:
        uint32_t length
        uint32_t padding1
        symbol_t* data

    cdef cppclass CGappedSequence:
        symbol_t* symbols
        size_t size
        size_t symbols_size
        size_t gapped_size
        size_t dps_size
        size_t dps_size_div2

        vector[uint32_t] n_gaps
        vector[uint32_t] dps

        string id
        vector[bool] uppercase

        CGappedSequence(CSequence&& sequence) except +
        CGappedSequence(const CGappedSequence&& sequence) except +
        CGappedSequence(CGappedSequence&& _gapped_sequence)

        void InsertGap(uint32_t pos) except +
        void InsertGaps(uint32_t pos, uint32_t n) except +
        void InsertGapsVector(const vector[pair[uint32_t, uint32_t]]& v_gaps) except +

        void RemoveGap(size_t pos) except +
        void RemoveGaps(size_t pos, uint32_t n) except +
        symbol_t GetSymbol(size_t pos)

        void DecodeRaw(symbol_t* seq) except +
        string Decode() except +
        uint32_t NoSymbols() except +

        void InsertFront(symbol_t new_symbol) except +

        void Clear() except +
        void ClearDPS() except +



import io
import os
import sys
import unittest
import textwrap
import tempfile

from .. import Aligner, Sequence
from . import fasta, data

try:
    try:
        import importlib.resources as importlib_resources
    except ImportError:
        import importlib_resources
except ImportError:
    importlib_resources = None


class _Test(object):

    @unittest.skipUnless(importlib_resources, "tests require `importlib.resources`")
    def test_hemopexin_medoid_nj(self):
        self._test_famsa("hemopexin", "nj")

    @unittest.skipUnless(importlib_resources, "tests require `importlib.resources`")
    def test_hemopexin_medoid_sl(self):
        self._test_famsa("hemopexin", "sl")

    @unittest.skipUnless(importlib_resources, "tests require `importlib.resources`")
    def test_hemopexin_medoid_upgma(self):
        self._test_famsa("hemopexin", "upgma")

    @unittest.skipUnless(importlib_resources, "tests require `importlib.resources`")
    def test_adeno_fiber_medoid_sl(self):
        self._test_famsa("adeno_fiber", "sl")

    @unittest.skipUnless(importlib_resources, "tests require `importlib.resources`")
    def test_adeno_fiber_medoid_upgma(self):
        self._test_famsa("adeno_fiber", "upgma")

    @unittest.skipUnless(importlib_resources, "tests require `importlib.resources`")
    def test_adeno_fiber_sl(self):
        self._test_famsa("adeno_fiber", "sl", None)

    @unittest.skipUnless(importlib_resources, "tests require `importlib.resources`")
    def test_adeno_fiber_upgma(self):
        self._test_famsa("adeno_fiber", "upgma", None)


class TestAlign(unittest.TestCase, _Test):

    def _test_famsa(self, test_case, guide_tree, tree_heuristic="medoid"):
        filename = "{}.faa".format(test_case)
        with importlib_resources.open_text(data.__name__, filename) as file:
            records = list(fasta.parse(file))

        if tree_heuristic is None:
            filename = "{}.{}.afa".format(test_case, guide_tree)
        else:
            filename = "{}.{}-{}.afa".format(test_case, tree_heuristic, guide_tree)
        with importlib_resources.open_text(data.__name__, filename) as file:
            result = list(fasta.parse(file))

        aligner = Aligner(guide_tree=guide_tree, tree_heuristic=tree_heuristic, threads=1)
        sequences = (
            Sequence(record.id.encode(), record.seq.encode())
            for record in records
        )

        alignment = aligner.align(sequences)
        for expected, actual in zip(result, alignment):
            self.assertEqual(expected.id, actual.id.decode())
            self.assertEqual(expected.seq, actual.sequence.decode())


class TestBuildTree(unittest.TestCase, _Test):

    def _test_famsa(self, test_case, guide_tree, tree_heuristic="medoid"):
        filename = "{}.faa".format(test_case)
        with importlib_resources.open_text(data.__name__, filename) as file:
            records = list(fasta.parse(file))

        if tree_heuristic is None:
            filename = "{}.{}.nwk".format(test_case, guide_tree)
        else:
            filename = "{}.{}-{}.nwk".format(test_case, tree_heuristic, guide_tree)
        with importlib_resources.open_text(data.__name__, filename) as file:
            result = file.read()

        aligner = Aligner(guide_tree=guide_tree, tree_heuristic=tree_heuristic, threads=1)
        sequences = (
            Sequence(record.id.encode(), record.seq.encode())
            for record in records
        )

        tree = aligner.build_tree(sequences)
        self.assertMultiLineEqual(
            tree.dumps().decode(),
            result,
        )



#!/usr/bin/env python3

import configparser
import functools
import glob
import itertools
import io
import multiprocessing.pool
import os
import platform
import re
import setuptools
import setuptools.extension
import subprocess
import sys
from distutils.command.clean import clean as _clean
from distutils.errors import CompileError
from setuptools.extension import Extension
from setuptools.command.build_clib import build_clib as _build_clib
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.sdist import sdist as _sdist

try:
    from Cython.Build import cythonize
except ImportError as err:
    cythonize = err

try:
    import semantic_version
except ImportError as err:
    semantic_version = err

# --- Constants -----------------------------------------------------------------

SETUP_FOLDER = os.path.relpath(os.path.realpath(os.path.join(__file__, os.pardir)))
INCLUDE_FOLDER = os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src")

MACHINE = platform.machine()
if re.match("^mips", MACHINE):
    TARGET_CPU = "mips"
elif re.match("^(aarch64|arm64)$", MACHINE):
    TARGET_CPU = "aarch64"
elif re.match("^arm", MACHINE):
    TARGET_CPU = "arm"
elif re.match("(x86_64)|(AMD64|amd64)|(^i.86$)", MACHINE):
    TARGET_CPU = "x86"
elif re.match("^(powerpc|ppc)", MACHINE):
    TARGET_CPU = "ppc"
else:
    TARGET_CPU = None

SYSTEM  = platform.system()
if SYSTEM == "Linux" or SYSTEM == "Java":
    TARGET_SYSTEM = "linux_or_android"
elif SYSTEM.endswith("FreeBSD"):
    TARGET_SYSTEM = "freebsd"
elif SYSTEM == "Darwin":
    TARGET_SYSTEM = "macos"
elif SYSTEM.startswith(("Windows", "MSYS", "MINGW", "CYGWIN")):
    TARGET_SYSTEM = "windows"
else:
    TARGET_SYSTEM = None

# --- Utils ------------------------------------------------------------------

_HEADER_PATTERN = re.compile("^@@ -(\d+),?(\d+)? \+(\d+),?(\d+)? @@$")


def _eprint(*args, **kwargs):
    print(*args, **kwargs, file=sys.stderr)


def _patch_osx_compiler(compiler):
    # On newer OSX, Python has been compiled as a universal binary, so
    # it will attempt to pass universal binary flags when building the
    # extension. This will not work because the code makes use of SSE2.
    for tool in ("compiler", "compiler_so", "linker_so"):
        flags = getattr(compiler, tool)
        i = next((i for i in range(1, len(flags)) if flags[i-1] == "-arch" and flags[i] != platform.machine()), None)
        if i is not None:
            flags.pop(i)
            flags.pop(i-1)
        if "-fPIC" not in flags:
            flags.append("-fPIC")
        while "-O3" in flags:
            i = flags.index("-O3")
            flags[i] = "-O2"


def _apply_patch(s,patch,revert=False):
    # see https://stackoverflow.com/a/40967337
    s = s.splitlines(keepends=True)
    p = patch.splitlines(keepends=True)
    t = []
    i = 0
    sl = 0
    midx, sign = (1,'+') if not revert else (3,'-')
    while i < len(p) and p[i].startswith(("---","+++")):
        i += 1 # skip header lines

    while i < len(p):
        match = _HEADER_PATTERN.match(p[i])
        if not match:
          raise ValueError("Invalid line in patch: {!r}".format(p[i]))
        i += 1
        l = int(match.group(midx)) - 1 + (match.group(midx+1) == '0')
        t.extend(s[sl:l])
        sl = l
        while i < len(p) and p[i][0] != '@':
            if i+1 < len(p) and p[i+1][0] == '\\':
                line = p[i][:-1]
                i += 2
            else:
                line = p[i]
                i += 1
            if len(line) > 0:
                if line[0] == sign or line[0] == ' ':
                    t += line[1:]
                sl += (line[0] != sign)

    t.extend(s[sl:])
    return ''.join(t)


def _is_compiler_clang(compiler):
    try:
        if compiler.compiler_type == "unix":
            proc = subprocess.run([compiler.compiler[0], "--version"], capture_output=True)
            if proc.returncode == 0:
                return any(
                    line.startswith(b"Apple clang")
                    for line in proc.stdout.splitlines()
                )
    except:
        pass
    return False


def _detect_compiler_version(compiler):
    try:
        if compiler.compiler_type == "unix":
            proc = subprocess.run([compiler.compiler[0], "-dumpversion"], capture_output=True)
            if proc.returncode == 0:
                return semantic_version.Version(proc.stdout.decode())
    except:
        pass
    return semantic_version.Version("5.0.0") # assume lowest compatibility


def _set_cpp_flags(compiler, extension):
    # see `vendor/FAMSA/makefile`
    cc_version = _detect_compiler_version(compiler)

    # check we are not using the buggy clang version
    if _is_compiler_clang(compiler) and cc_version.major == 13 and cc_version.minor == 0:
        raise RuntimeError(
            "clang 13.0 is known to produce undefined behaviour "
            "(see https://github.com/refresh-bio/FAMSA/issues/32), "
            "please update or use a different compiler."
        )

    if cc_version.major <= 6:
        cpp_target = "11"
        defines = [("NO_PROFILE_PAR", 1), ("OLD_ATOMIC_FLAG", 1)]
    elif cc_version.major == 7:
        cpp_target = "14"
        defines = [("NO_PROFILE_PAR", 1), ("OLD_ATOMIC_FLAG", 1)]
    elif cc_version.major <= 10:
        cpp_target = "2a"
        defines = [("OLD_ATOMIC_FLAG", 1)]
    else:
        cpp_target = "20"
        defines = []

    extension.define_macros.extend(defines)
    if compiler.compiler_type in {"unix", "cygwin", "mingw32"}:
        extension.extra_compile_args.append("-std=c++{}".format(cpp_target))
        extension.extra_compile_args.append("-pthread")
        extension.extra_compile_args.append("-Wno-alloc-size-larger-than")
        extension.extra_compile_args.append("-Wno-char-subscripts")
        extension.extra_link_args.append("-pthread")
    elif compiler.compiler_type == "msvc":
        extension.extra_compile_args.append("/std:c{}".format(cpp_target))


def _set_debug_flags(compiler, extension):
    if compiler.compiler_type in {"unix", "cygwin", "mingw32"}:
        extension.extra_compile_args.append("-g")
    elif compiler.compiler_type == "msvc":
        extension.extra_compile_args.append("/Z7")


# --- Extension with SIMD support --------------------------------------------

class Library(setuptools.extension.Library):

    def __init__(self, *args, **kwargs):
        self._needs_stub = False
        self.platform_sources = kwargs.pop("platform_sources", {})
        super().__init__(*args, **kwargs)


# --- Commands ------------------------------------------------------------------

class sdist(_sdist):
    """A `sdist` that generates a `pyproject.toml` on the fly.
    """

    def run(self):
        # build `pyproject.toml` from `setup.cfg`
        c = configparser.ConfigParser()
        c.add_section("build-system")
        c.set("build-system", "requires", str(self.distribution.setup_requires))
        c.set("build-system", 'build-backend', '"setuptools.build_meta"')
        with open("pyproject.toml", "w") as pyproject:
            c.write(pyproject)
        # run the rest of the packaging
        _sdist.run(self)


class build_ext(_build_ext):
    """A `build_ext` that disables optimizations if compiled in debug mode.
    """

    # --- Compatibility with `setuptools.Command`

    user_options = _build_ext.user_options + [
        ("disable-avx", None, "Force compiling the extension without AVX instructions"),
        ("disable-avx2", None, "Force compiling the extension without AVX2 instructions"),
        ("disable-neon", None, "Force compiling the extension without NEON instructions"),
    ]

    def initialize_options(self):
        _build_ext.initialize_options(self)
        self.disable_avx = False
        self.disable_avx2 = TARGET_SYSTEM == "macos"
        self.disable_neon = False

    def finalize_options(self):
        _build_ext.finalize_options(self)
        # transfer arguments to the build_clib method
        self._clib_cmd = self.get_finalized_command("build_clib")
        self._clib_cmd.debug = self.debug
        self._clib_cmd.force = self.force
        self._clib_cmd.verbose = self.verbose
        self._clib_cmd.define = self.define
        self._clib_cmd.include_dirs = self.include_dirs
        self._clib_cmd.compiler = self.compiler
        self._clib_cmd.parallel = self.parallel
        self._clib_cmd.disable_avx = self.disable_avx
        self._clib_cmd.disable_avx2 = self.disable_avx2
        self._clib_cmd.disable_neon = self.disable_neon

    # --- Build code ---

    def build_extension(self, ext):
        # show the compiler being used
        _eprint("building", ext.name, "with", self.compiler.compiler_type, "compiler")

        # add debug symbols if we are building in debug mode
        if self.debug:
            _set_debug_flags(self.compiler, ext)
            if sys.implementation.name == "cpython":
                ext.define_macros.append(("CYTHON_TRACE_NOGIL", 1))
        else:
            ext.define_macros.append(("CYTHON_WITHOUT_ASSERTIONS", 1))

        # add C++11 flags
        _set_cpp_flags(self.compiler, ext)

        # add Windows flags
        if TARGET_SYSTEM == "windows":
            if self.compiler.compiler_type == "msvc":
                ext.define_macros.append(("WIN32", 1))

        # update link and include directories
        for name in ext.libraries:
            lib = self._clib_cmd.get_library(name)
            if lib is not None:
                libfile = self.compiler.library_filename(
                    lib.name, output_dir=self._clib_cmd.build_clib
                )
                ext.depends.append(libfile)
                ext.extra_objects.append(libfile)
                ext.extra_objects.extend(lib.extra_objects)
                ext.define_macros.extend(lib.define_macros)

        # build the rest of the extension as normal
        ext._needs_stub = False
        _build_ext.build_extension(self, ext)

    def build_extensions(self):
        # check `cythonize` is available
        if isinstance(cythonize, ImportError):
            raise RuntimeError("Cython is required to run `build_ext` command") from cythonize

        # remove universal compilation flags for OSX
        if platform.system() == "Darwin":
            _patch_osx_compiler(self.compiler)

        # use debug directives with Cython if building in debug mode
        cython_args = {
            "include_path": ["include"],
            "compiler_directives": {
                "cdivision": True,
                "nonecheck": False,
            },
            "compile_time_env": {
                "SYS_IMPLEMENTATION_NAME": sys.implementation.name,
                "SYS_VERSION_INFO_MAJOR": sys.version_info.major,
                "SYS_VERSION_INFO_MINOR": sys.version_info.minor,
                "SYS_VERSION_INFO_MICRO": sys.version_info.micro,
                "DEFAULT_BUFFER_SIZE": io.DEFAULT_BUFFER_SIZE,
                "TARGET_CPU": TARGET_CPU,
                "TARGET_SYSTEM": TARGET_SYSTEM,
            }
        }
        if self.force:
            cython_args["force"] = True
        if self.debug:
            cython_args["annotate"] = True
            cython_args["compiler_directives"]["cdivision_warnings"] = True
            cython_args["compiler_directives"]["warn.undeclared"] = True
            cython_args["compiler_directives"]["warn.unreachable"] = True
            cython_args["compiler_directives"]["warn.maybe_uninitialized"] = True
            cython_args["compiler_directives"]["warn.unused"] = True
            cython_args["compiler_directives"]["warn.unused_arg"] = True
            cython_args["compiler_directives"]["warn.unused_result"] = True
            cython_args["compiler_directives"]["warn.multiple_declarators"] = True
        else:
            cython_args["compiler_directives"]["boundscheck"] = False
            cython_args["compiler_directives"]["wraparound"] = False

        # compile the C library
        if not self.distribution.have_run.get("build_clib", False):
            self._clib_cmd.run()

        # add the include dirs from the `build_clib` command
        for ext in self.extensions:
            ext.include_dirs.insert(0, self._clib_cmd.build_clib)

        # cythonize the extensions and build normally
        self.extensions = cythonize(self.extensions, **cython_args)
        _build_ext.build_extensions(self)


class build_clib(_build_clib):
    """A custom `build_clib` that makes all C++ class attributes public.
    """

    # --- Compatibility with `setuptools.Command`

    user_options = _build_clib.user_options + [
        ("parallel=", "j", "number of parallel build jobs"),
        ("disable-avx", None, "Force compiling the library without AVX instructions"),
        ("disable-avx2", None, "Force compiling the library without AVX2 instructions"),
        ("disable-neon", None, "Force compiling the library without NEON instructions"),
    ]

    def initialize_options(self):
        _build_clib.initialize_options(self)
        self.parallel = None
        self.disable_avx = False
        self.disable_avx2 = TARGET_SYSTEM == "macos"
        self.disable_neon = False

    def finalize_options(self):
        _build_clib.finalize_options(self)
        if self.parallel is not None:
            self.parallel = int(self.parallel)
        # record SIMD-specific options
        self._simd_supported = dict(AVX=False, AVX2=False, NEON=False)
        self._simd_defines = dict(AVX=[], AVX2=[], NEON=[])
        self._simd_flags = dict(AVX=[], AVX2=[], NEON=[])
        self._simd_disabled = {
            "AVX": self.disable_avx,
            "AVX2": self.disable_avx2,
            "NEON": self.disable_neon,
        }

    # --- Autotools-like helpers ---

    def _patch_file(self, input, output):
        basename = os.path.basename(input)
        patchname = os.path.realpath(os.path.join(__file__, os.pardir, "patches", "{}.patch".format(basename)))
        if os.path.exists(patchname):
            _eprint("patching", os.path.relpath(input), "with", os.path.relpath(patchname))
            with open(patchname, "r") as patchfile:
                patch = patchfile.read()
            with open(input, "r") as src:
                srcdata = src.read()
            with open(output, "w") as dst:
                dst.write(_apply_patch(srcdata, patch))
        else:
            self.copy_file(input, output)

    def _check_function(self, funcname, header, args="()"):
        _eprint('checking whether function', repr(funcname), 'is available', end="... ")

        self.mkpath(self.build_temp)
        base = "have_{}".format(funcname)
        testfile = os.path.join(self.build_temp, "{}.c".format(base))
        binfile = self.compiler.executable_filename(base, output_dir=self.build_temp)
        objects = []

        with open(testfile, "w") as f:
            f.write("""
                #include <{}>
                int main() {{
                    {}{};
                    return 0;
                }}
            """.format(header, funcname, args))
        try:
            objects = self.compiler.compile([testfile], debug=self.debug)
            self.compiler.link_executable(objects, base, output_dir=self.build_temp)
        except CompileError:
            _eprint("no")
            return False
        else:
            _eprint("yes")
            return True
        finally:
            os.remove(testfile)
            for obj in filter(os.path.isfile, objects):
                os.remove(obj)
            if os.path.isfile(binfile):
                os.remove(binfile)

    def _check_simd_generic(self, name, flags, header, vector, set, op, extract, result=1):
        _eprint('checking whether compiler can build', name, 'code', end="... ")

        base = "have_{}".format(name)
        testfile = os.path.join(self.build_temp, "{}.c".format(base))
        binfile = self.compiler.executable_filename(base, output_dir=self.build_temp)
        objects = []

        self.mkpath(self.build_temp)
        with open(testfile, "w") as f:
            f.write("""
                #include <{}>
                int main() {{
                    {}      a = {}(1);
                            a = {}(a);
                    short   x = {}(a, 1);
                    return (x == 1) ? 0 : 1;
                }}
            """.format(header, vector, set, op, extract))

        try:
            self.mkpath(self.build_temp)
            objects = self.compiler.compile([testfile], extra_preargs=flags)
            self.compiler.link_executable(objects, base, output_dir=self.build_temp)
            subprocess.run([binfile], check=True)
        except CompileError:
            _eprint("no")
            return False
        except subprocess.CalledProcessError:
            _eprint("yes, but cannot run code")
            return True  # assume we are cross-compiling, and still build
        else:
            if not flags:
                _eprint("yes")
            else:
                _eprint("yes, with {}".format(" ".join(flags)))
            return True
        finally:
            os.remove(testfile)
            for obj in filter(os.path.isfile, objects):
                os.remove(obj)
            if os.path.isfile(binfile):
                os.remove(binfile)

    def _avx_flags(self):
        if self.compiler.compiler_type == "msvc":
            return ["/arch:AVX2"]
        return ["-mavx", "-mpopcnt", "-funroll-loops"]

    def _check_avx(self):
        return self._check_simd_generic(
            "AVX",
            self._avx_flags(),
            header="immintrin.h",
            vector="__m256i",
            set="_mm256_set1_epi16",
            op="",
            extract="_mm256_extract_epi16",
        )

    def _avx2_flags(self):
        if self.compiler.compiler_type == "msvc":
            return ["/arch:AVX2"]
        return ["-mavx2", "-mpopcnt", "-funroll-loops"]

    def _check_avx2(self):
        return self._check_simd_generic(
            "AVX2",
            self._avx2_flags(),
            header="immintrin.h",
            vector="__m256i",
            set="_mm256_set1_epi16",
            op="_mm256_abs_epi32",
            extract="_mm256_extract_epi16",
        )

    def _neon_flags(self):
        return ["-mfpu=neon"] if TARGET_CPU == "arm" else []

    def _check_neon(self):
        return self._check_simd_generic(
            "NEON",
            self._neon_flags(),
            header="arm_neon.h",
            vector="int16x8_t",
            set="vdupq_n_s16",
            op="vabsq_s16",
            extract="vgetq_lane_s16"
        )

    # --- Compatibility with base `build_clib` command ---

    def check_library_list(self, libraries):
        pass

    def get_library_names(self):
        return [ lib.name for lib in self.libraries ]

    def get_source_files(self):
        return [ source for lib in self.libraries for source in lib.sources ]

    def get_library(self, name):
        return next((lib for lib in self.libraries if lib.name == name), None)

    # --- Build code ---

    def build_simd_code(self, library):
        # build platform-specific code
        for simd, sources in library.platform_sources.items():
            if self._simd_supported[simd] and not self._simd_disabled[simd]:
                objects = [
                    os.path.join(self.build_temp, s.replace(".cpp", self.compiler.obj_extension))
                    for s in sources
                ]
                for source, object in zip(sources, objects):
                    self.make_file(
                        [source],
                        object,
                        self.compiler.compile,
                        (
                            [source],
                            self.build_temp,
                            library.define_macros + self._simd_defines[simd],
                            library.include_dirs,
                            self.debug,
                            library.extra_compile_args + self._simd_flags[simd],
                            None,
                            library.depends
                        )
                    )
                library.extra_objects.extend(objects)

    def build_libraries(self, libraries):
        # check `semantic_version` is available
        if isinstance(semantic_version, ImportError):
            raise RuntimeError("`semantic_version` is required to run `build_ext` command") from semantic_version

        # check for functions required for libcpu_features on OSX
        if TARGET_SYSTEM == "macos":
            _patch_osx_compiler(self.compiler)

        # check if we can build platform-specific code
        if TARGET_CPU == "x86":
            if not self._simd_disabled["AVX"] and self._check_avx():
                self._simd_supported["AVX"] = True
                self._simd_flags["AVX"].extend(self._avx_flags())
            if not self._simd_disabled["AVX2"] and self._check_avx2():
                self._simd_supported["AVX2"] = True
                self._simd_flags["AVX2"].extend(self._avx2_flags())
        elif TARGET_CPU == "arm" or TARGET_CPU == "aarch64":
            if not self._simd_disabled["NEON"] and self._check_neon():
                self._simd_supported["NEON"] = True
                self._simd_flags["NEON"].extend(self._neon_flags())

        # setup dispatcher of SIMD extensions
        for library in libraries:
            if self._simd_supported["AVX2"]:
                library.define_macros.append(("SIMD", 2))
            elif self._simd_supported["AVX"]:
                library.define_macros.append(("SIMD", 1))
            elif self._simd_supported["NEON"]:
                library.define_macros.append(("SIMD", 4))
            else:
                library.define_macros.append(("SIMD", 0))

        # build each library only if the sources are outdated
        self.mkpath(self.build_clib)
        for library in libraries:
            libname = self.compiler.library_filename(library.name, output_dir=self.build_clib)
            # self.make_file(library.sources, libname, self.build_library, (library,))
            self.build_library(library)

    def build_library(self, library):
        # show the compiler being used
        _eprint("building", library.name, "with", self.compiler.compiler_type, "compiler")

        # add debug flags if we are building in debug mode
        if self.debug:
            _set_debug_flags(self.compiler, library)
        # add C++11 flags
        if library.name == "famsa":
            _set_cpp_flags(self.compiler, library)

        # add Windows flags
        if TARGET_SYSTEM == "windows" and self.compiler.compiler_type == "msvc":
            library.define_macros.append(("WIN32", 1))

        # copy headers w/ patches to build directory
        for header in library.depends:
            output = os.path.join(self.build_clib, os.path.relpath(header, INCLUDE_FOLDER))
            self.mkpath(os.path.dirname(output))
            self.make_file(
                [header],
                output,
                self._patch_file,
                (header, output)
            )

        # fix include dirs to use the build directory folders
        for i, include_dir in enumerate(library.include_dirs):
            if include_dir.startswith(INCLUDE_FOLDER):
                library.include_dirs[i] = os.path.join(self.build_clib, os.path.relpath(include_dir, INCLUDE_FOLDER))

        # copy sources w/ patches to build directory
        sources = [
            os.path.join(self.build_temp, os.path.basename(source))
            for source in library.sources
        ]
        for source, source_copy in zip(library.sources, sources):
            self.make_file(
                [source],
                source_copy,
                self._patch_file,
                (source, source_copy)
            )

        # store compile args
        compile_args = (
            None,
            library.define_macros,
            library.include_dirs + [self.build_clib],
            self.debug,
            library.extra_compile_args,
            None,
            library.depends,
        )
        # manually prepare sources and get the names of object files
        objects = [
            re.sub(r'(.cpp|.c)$', self.compiler.obj_extension, s)
            for s in sources
        ]
        # only compile outdated files
        with multiprocessing.pool.ThreadPool(self.parallel) as pool:
            pool.starmap(
                functools.partial(self._compile_file, compile_args=compile_args),
                zip(sources, objects)
            )

        # build platform-specific code
        self.build_simd_code(library)

        # link into a static library
        libfile = self.compiler.library_filename(
            library.name,
            output_dir=self.build_clib,
        )
        self.make_file(
            objects,# + library.extra_objects,
            libfile,
            self.compiler.create_static_lib,
            (objects, library.name, self.build_clib, None, self.debug)
        )

    def _compile_file(self, source, object, compile_args):
        self.make_file(
            [source],
            object,
            self.compiler.compile,
            ([source], *compile_args)
        )


class clean(_clean):
    """A `clean` that removes intermediate files created by Cython.
    """

    def run(self):

        source_dir = os.path.join(os.path.dirname(__file__), "pyfamsa")

        patterns = ["*.html"]
        if self.all:
            patterns.extend(["*.so", "*.c", "*.cpp"])

        for pattern in patterns:
            for file in glob.glob(os.path.join(source_dir, pattern)):
                _eprint("removing {!r}".format(file))
                os.remove(file)

        _clean.run(self)


# --- Setup ---------------------------------------------------------------------

setuptools.setup(
    libraries=[
        # NOTE(@althonos): libdeflate is only needed for `IOService`, but we
        #                  don't use the FAMSA I/O so we don't have to build
        #                  it and link to it, as long as we're not building
        #                  `io_service.cpp` either
        Library(
            "famsa",
            language="c++",
            sources=[
                # COMMON_OBJS
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "msa.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","AbstractTreeGenerator.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","Clustering.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","DistanceCalculator.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","FastTree.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","GuideTree.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","MSTPrim.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","NeighborJoining.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","NewickParser.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","SingleLinkage.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree","UPGMA.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "timer.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "log.cpp"),
                # os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "io_service.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "params.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "profile.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "profile_par.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "profile_seq.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "sequence.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "queues.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "libs", "mimalloc", "static.cpp"),
                # LCS_OBJS
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp.cpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp_classic.cpp"),
            ],
            include_dirs=[
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "libs", "mimalloc"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "libs"),
                os.path.join(SETUP_FOLDER, "vendor"),
                os.path.join(SETUP_FOLDER, "include"),
            ],
            depends=[
                # core
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "defs.h"),
                # os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "io_service.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "params.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "profile.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "queues.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "sequence.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "core", "version.h"),
                # lcs
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp_avx_intr.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp_avx2_intr.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp_classic.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp_neon_intr.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp.h"),
                # tree
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "AbstractTreeGenerator.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "AbstractTreeGenerator.hpp"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "Chained.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "Clustering.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "DistanceCalculator.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "FastTree.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "GuideTree.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "IPartialGenerator.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "MSTPrim.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "NeighborJoining.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "NewickParser.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "SingleLinkage.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "SingleLinkageQueue.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "TreeDefs.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "TreeGrammar.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "tree", "UPGMA.h"),
                # utils
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "array.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "conversion.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "cpuid.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "deterministic_random.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "log.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "memory_monotonic.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "meta_oper.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "pooled_threads.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "statistics.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "timer.h"),
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "utils.h"),
                # msa
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "msa.h"),
            ],
            platform_sources={
                "AVX": [
                    os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "utils_avx.cpp"),
                    os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp_avx_intr.cpp"),
                ],
                "AVX2": [
                    os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "utils_avx2.cpp"),
                    os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp_avx2_intr.cpp"),
                ],
                "NEON": [
                    os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "utils_neon.cpp"),
                    os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "lcs", "lcsbp_neon_intr.cpp"),
                ]
            },
        ),
    ],
    ext_modules=[
        Extension(
            "pyfamsa._famsa",
            language="c++",
            sources=[
                # for some reason, if this file is not compiled in the extension
                # this causes the `clear_mem` symbol to be missing from the
                # final shared object
                os.path.join(SETUP_FOLDER, "vendor", "FAMSA", "src", "utils", "utils.cpp"),
                os.path.join(SETUP_FOLDER, "pyfamsa", "_famsa.pyx"),
            ],

            include_dirs=[
                os.path.join(SETUP_FOLDER, "include"),
            ],
            libraries=[
                "famsa",
            ],
        ),
    ],
    cmdclass={
        "sdist": sdist,
        "build_ext": build_ext,
        "build_clib": build_clib,
        "clean": clean
    }
)



from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector

from famsa.core.sequence cimport CSequence, CGappedSequence
from famsa.utils.memory_monotonic cimport memory_monotonic_safe

cdef extern from "core/io_service.h" nogil:

    cdef cppclass IOService:
        @staticmethod
        size_t loadFasta(const string& filename, vector[CSequence]& sequences) except +
        # @staticmethod
        # size_t loadFasta(const string& filename, vector[CSequence]& sequences, memory_monotonic_safe* mma) except +
        @staticmethod
        bool saveAlignment(const string& filename, vector[CGappedSequence*]& sequences, int no_threads, int gzip_level) except +
